from collections import namedtuple

from pkg_resources import iter_entry_points

from flake8_import_order import (
    ClassifiedImport, IMPORT_3RD_PARTY, IMPORT_APP, IMPORT_APP_RELATIVE,
    NewLine,
)

Error = namedtuple('Error', ['lineno', 'code', 'message'])


def list_entry_points():
    return iter_entry_points('flake8_import_order.styles')


def lookup_entry_point(name):
    try:
        return next(iter_entry_points('flake8_import_order.styles', name=name))
    except StopIteration:
        raise LookupError('Unknown style {}'.format(name))


class Style(object):

    accepts_application_package_names = False

    def __init__(self, nodes):
        self.nodes = nodes

    def check(self):
        previous = None
        previous_import = None
        for current in self.nodes:
            if isinstance(current, ClassifiedImport):
                for error in self._check(previous_import, previous, current):
                    yield error
                previous_import = current
            previous = current

    def _check(self, previous_import, previous, current_import):
        if current_import.type == -1:
            yield Error(
                current_import.lineno,
                'I666',
                'Import statement mixes groups',
            )

        correct_names = self.sorted_names(current_import.names)
        if correct_names != current_import.names:
            corrected = ', '.join(correct_names)
            yield Error(
                current_import.lineno,
                'I101',
                "Imported names are in the wrong order. "
                "Should be {0}".format(corrected),
            )

        if previous_import is not None:
            previous_key = self.import_key(previous_import)
            current_key = self.import_key(current_import)
            if previous_key > current_key:
                first = self._explain(current_import)
                second = self._explain(previous_import)
                if first == second:
                    first = ", ".join(current_import.names)
                    second = ", ".join(previous_import.names)
                yield Error(
                    current_import.lineno,
                    'I100',
                    "Import statements are in the wrong order. "
                    "{0} should be before {1}".format(first, second),
                )

            same_section = self.same_section(previous_import, current_import)
            has_newline = isinstance(previous, NewLine)
            if not same_section and not has_newline:
                yield Error(
                    current_import.lineno,
                    'I201',
                    'Missing newline before sections or imports.',
                )

            if same_section and has_newline:
                yield Error(
                    current_import.lineno,
                    'I202',
                    'Additional newline in a section of imports.',
                )

    @staticmethod
    def sorted_names(names):
        return names

    @staticmethod
    def import_key(import_):
        return (import_.type,)

    @staticmethod
    def same_section(previous, current):
        same_type = current.type == previous.type
        both_first = (
            {previous.type, current.type} <= {IMPORT_APP, IMPORT_APP_RELATIVE}
        )
        return same_type or both_first

    @staticmethod
    def _explain(import_):
        if import_.is_from:
            text = 'from ' + import_.level * '.'
        else:
            text = 'import '
        return text + ', '.join(import_.modules)


class PEP8(Style):
    pass


class Google(Style):

    @staticmethod
    def sorted_names(names):
        return sorted(names, key=Google.name_key)

    @staticmethod
    def name_key(name):
        return (name.lower(), name)

    @staticmethod
    def import_key(import_):
        modules = [Google.name_key(module) for module in import_.modules]
        names = [Google.name_key(name) for name in import_.names]
        return (import_.type, import_.level, modules, names)


class AppNexus(Google):
    accepts_application_package_names = True


class Smarkets(Style):

    @staticmethod
    def sorted_names(names):
        return sorted(names, key=Smarkets.name_key)

    @staticmethod
    def name_key(name):
        return (name.lower(), name)

    @staticmethod
    def import_key(import_):
        modules = [Smarkets.name_key(module) for module in import_.modules]
        names = [Smarkets.name_key(name) for name in import_.names]
        return (import_.type, import_.is_from, import_.level, modules, names)


class Edited(Smarkets):
    accepts_application_package_names = True


class Cryptography(Style):

    @staticmethod
    def sorted_names(names):
        return sorted(names)

    @staticmethod
    def import_key(import_):
        if import_.type in {IMPORT_3RD_PARTY, IMPORT_APP}:
            return (
                import_.type, import_.package, import_.is_from,
                import_.level, import_.modules, import_.names,
            )
        else:
            return (
                import_.type, '', import_.is_from, import_.level,
                import_.modules, import_.names,
            )

    @staticmethod
    def same_section(previous, current):
        app_or_third = current.type in {IMPORT_3RD_PARTY, IMPORT_APP}
        same_type = current.type == previous.type
        both_relative = previous.type == current.type == IMPORT_APP_RELATIVE
        same_package = previous.package == current.package
        return (
            (not app_or_third and same_type or both_relative) or
            (app_or_third and same_package)
        )



import ast
from collections import namedtuple

from flake8_import_order.__about__ import (
    __author__, __copyright__, __email__, __license__, __summary__, __title__,
    __uri__, __version__,
)
from flake8_import_order.stdlib_list import STDLIB_NAMES

__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

DEFAULT_IMPORT_ORDER_STYLE = 'cryptography'

IMPORT_FUTURE = 0
IMPORT_STDLIB = 10
IMPORT_3RD_PARTY = 20
IMPORT_APP_PACKAGE = 30
IMPORT_APP = 40
IMPORT_APP_RELATIVE = 50
IMPORT_MIXED = -1

ClassifiedImport = namedtuple(
    'ClassifiedImport',
    ['type', 'is_from', 'modules', 'names', 'lineno', 'level', 'package'],
)
NewLine = namedtuple('NewLine', ['lineno'])


def get_package_names(name):
    tree = ast.parse(name)
    parts = []
    for node in ast.walk(tree):
        if isinstance(node, ast.Attribute):
            parts.append(node.attr)
        if isinstance(node, ast.Name):
            parts.append(node.id)

    if not parts:
        return []

    last_package_name = parts.pop()
    package_names = [last_package_name]

    for part in reversed(parts):
        last_package_name = '%s.%s' % (last_package_name, part)
        package_names.append(last_package_name)

    return package_names


def root_package_name(name):
    tree = ast.parse(name)
    for node in ast.walk(tree):
        if isinstance(node, ast.Name):
            return node.id
    else:
        return None


class ImportVisitor(ast.NodeVisitor):

    def __init__(self, application_import_names, application_package_names):
        self.imports = []
        self.application_import_names = frozenset(application_import_names)
        self.application_package_names = frozenset(application_package_names)

    def visit_Import(self, node):  # noqa
        if node.col_offset == 0:
            modules = [alias.name for alias in node.names]
            types_ = {self._classify_type(module) for module in modules}
            if len(types_) == 1:
                type_ = types_.pop()
            else:
                type_ = IMPORT_MIXED
            classified_import = ClassifiedImport(
                type_, False, modules, [], node.lineno, 0,
                root_package_name(modules[0]),
            )
            self.imports.append(classified_import)

    def visit_ImportFrom(self, node):  # noqa
        if node.col_offset == 0:
            module = node.module or ''
            if node.level > 0:
                type_ = IMPORT_APP_RELATIVE
            else:
                type_ = self._classify_type(module)
            names = [alias.name for alias in node.names]
            classified_import = ClassifiedImport(
                type_, True, [module], names,
                node.lineno, node.level,
                root_package_name(module),
            )
            self.imports.append(classified_import)

    def _classify_type(self, module):
        package_names = get_package_names(module)

        # Walk through package names from most-specific to least-specific,
        # taking the first match found.
        for package in reversed(package_names):
            if package == "__future__":
                return IMPORT_FUTURE
            elif package in self.application_import_names:
                return IMPORT_APP
            elif package in self.application_package_names:
                return IMPORT_APP_PACKAGE
            elif package in STDLIB_NAMES:
                return IMPORT_STDLIB

        # Not future, stdlib or an application import.
        # Must be 3rd party.
        return IMPORT_3RD_PARTY
